/**
 * Copyright (c) 2014 - 2017, Nordic Semiconductor ASA
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 * 
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 * 
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 * 
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 * 
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/** @file
 *
 * @defgroup ble_sdk_app_beacon_main main.c
 * @{
 * @ingroup ble_sdk_app_beacon
 * @brief Beacon Transmitter Sample Application main file.
 *
 * This file contains the source code for an Beacon transmitter sample application.
 */

#include <stdbool.h>
#include <stdint.h>
#include "nordic_common.h"
#include "bsp.h"
#include "nrf_soc.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "ble_advdata.h"
#include "app_timer.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"


#define APP_BLE_CONN_CFG_TAG            1                                 /**< A tag identifying the SoftDevice BLE configuration. */

#define NON_CONNECTABLE_ADV_INTERVAL    MSEC_TO_UNITS(1000, UNIT_0_625_MS) /**< The advertising interval for non-connectable advertisement (100 ms). This value can vary between 100ms to 10.24s). */

#define APP_BEACON_INFO_LENGTH          0x17                              /**< Total length of information advertised by the Beacon. */
#define APP_ADV_DATA_LENGTH             0x15                              /**< Length of manufacturer specific data in the advertisement. */
#define APP_DEVICE_TYPE                 0x02                              /**< 0x02 refers to Beacon. */
#define APP_MEASURED_RSSI               0xC3                              /**< The Beacon's measured RSSI at 1 meter distance in dBm. */
#define APP_COMPANY_IDENTIFIER          0x0059                            /**< Company identifier for Nordic Semiconductor ASA. as per www.bluetooth.org. */
#define APP_MAJOR_VALUE                 0x01, 0x02                        /**< Major value used to identify Beacons. */
#define APP_MINOR_VALUE                 0x03, 0x04                        /**< Minor value used to identify Beacons. */
#define APP_BEACON_UUID                 0x01, 0x12, 0x23, 0x34, \
                                        0x45, 0x56, 0x67, 0x78, \
                                        0x89, 0x9a, 0xab, 0xbc, \
                                        0xcd, 0xde, 0xef, 0xf0            /**< Proprietary UUID for Beacon. */

#define DEAD_BEEF                       0xDEADBEEF                        /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

#define ADVERT_TIME                     APP_TIMER_TICKS(1000)
#define ADVERT_TIME_DELAY               APP_TIMER_TICKS(500)
#define SCAN_TIME                       APP_TIMER_TICKS(5000)
#define SCAN_TIME_DELAY                 APP_TIMER_TICKS(500)

APP_TIMER_DEF(m_adv_timer_id); 

#define SCAN_INTERVAL                   0x00A0//0x0004                                /**< Determines scan interval in units of 0.625 millisecond. */
#define SCAN_WINDOW                     0x00A0//0x0004                                /**< Determines scan window in units of 0.625 millisecond. */
#define SCAN_TIMEOUT                    0x0000                              /**< Timout when scanning. 0x0000 disables timeout. */
#define APP_BLE_OBSERVER_PRIO           3                                   /**< Application's BLE observer priority. You shouldn't need to modify this value. */

#define RELAY_MAIN_1_PIO                9
#define RELAY_MAIN_2_PIO                10
#define RELAY_RIGHT_3_PIO               26
#define RELAY_RIGHT_4_PIO               27
#define RELAY_LEFT_5_PIO                0
#define RELAY_LEFT_6_PIO                1

#define TEST_LED                        20

#if defined(USE_UICR_FOR_MAJ_MIN_VALUES)
#define MAJ_VAL_OFFSET_IN_BEACON_INFO   18                                /**< Position of the MSB of the Major Value in m_beacon_info array. */
#define UICR_ADDRESS                    0x10001080                        /**< Address of the UICR register used by this example. The major and minor versions to be encoded into the advertising data will be picked up from this location. */
#endif

/* File ID and Key used for the configuration record. */
#define CONFIG_FILE     (0xF010)
#define CONFIG_REC_KEY  (0x7010)

static uint32_t m_temp_scan_count;

typedef enum
{
    dev_state_advert,
    dev_state_advert_delay,
    dev_state_scan,
    dev_state_scan_delay
}dev_state_t;

typedef struct
{
    uint8_t relay_state;
    uint16_t ig_off_time_relay[6];
    bool    trigger_sw_state;
}relay_box_state_t;


static ble_gap_adv_params_t m_adv_params;                                 /**< Parameters to be passed to the stack when starting advertising. */

static dev_state_t m_dev_state;

static relay_box_state_t m_rbox_state;

/**@brief Parameters used when scanning. */
static ble_gap_scan_params_t const m_scan_params =
{
    .active   = 1,
    .interval = SCAN_INTERVAL,
    .window   = SCAN_WINDOW,
    .timeout  = SCAN_TIMEOUT,
    #if (NRF_SD_BLE_API_VERSION <= 2)
        .selective   = 0,
        .p_whitelist = NULL,
    #endif
    #if (NRF_SD_BLE_API_VERSION >= 3)
        .use_whitelist = 0,
    #endif
};

/**@brief Variable length data encapsulation in terms of length and pointer to data. */
typedef struct
{
    uint8_t * p_data;   /**< Pointer to data. */
    uint16_t  data_len; /**< Length of data. */
} data_t;


/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in]   line_num   Line number of the failing ASSERT call.
 * @param[in]   file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}

/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
static void advertising_init(void)
{
    uint32_t      err_code;
    ble_advdata_t advdata;
    uint8_t       flags = BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED;
    uint8_t manu_data[20];
   // ble_gap_addr_t device_addr;

    ble_advdata_manuf_data_t manuf_specific_data;

    manuf_specific_data.company_identifier = APP_COMPANY_IDENTIFIER;

    memset(&manu_data, 0, sizeof(manu_data));

    /* 4 bytes identifier for ReddShift  */
    manu_data[0] = 0x00; 
    manu_data[1] = 0x11; 
    manu_data[2] = 0x22; 
    manu_data[3] = 0x33; 

    /* 1 byte opcode Relay Box Beacon Message*/
    manu_data[4] = 0x02;

    /*1 byte relay state*/
    manu_data[5] = m_rbox_state.relay_state;

    /*1 byte Relays connected*/
    manu_data[6] = 0x3F;

    /* 1 byte ignition on signal */
    manu_data[7] = 1;

    uint8_t index;
    for(index=0; index<6; index++)
    {
        uint16_encode(m_rbox_state.ig_off_time_relay[index], &manu_data[8+2*index]);
    }

    /* 6 byte bt address
    err_code = sd_ble_gap_addr_get(&device_addr);
    manu_data[5] = device_addr.addr[5];
    manu_data[6] = device_addr.addr[4];
    manu_data[7] = device_addr.addr[3];
    manu_data[8] = device_addr.addr[2];
    manu_data[9] = device_addr.addr[1];
    manu_data[10] = device_addr.addr[0];
    */

    /* 1 byte bitmask of gear knob */
//    manu_data[11] = gear_state;

    manuf_specific_data.data.p_data  = (uint8_t *) manu_data;
    manuf_specific_data.data.size = 20;

    // Build and set advertising data.
    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type             = BLE_ADVDATA_NO_NAME;
    advdata.flags                 = flags;
    advdata.p_manuf_specific_data = &manuf_specific_data;

    err_code = ble_advdata_set(&advdata, NULL);
    APP_ERROR_CHECK(err_code);

    // Initialize advertising parameters (used when starting advertising).
    memset(&m_adv_params, 0, sizeof(m_adv_params));

    m_adv_params.type        = BLE_GAP_ADV_TYPE_ADV_NONCONN_IND;
    m_adv_params.p_peer_addr = NULL;    // Undirected advertisement.
    m_adv_params.fp          = BLE_GAP_ADV_FP_ANY;
    m_adv_params.interval    = NON_CONNECTABLE_ADV_INTERVAL;
    m_adv_params.timeout     = 0;       // Never time out.
}


/**@brief Function for starting advertising.
 */
static void advertising_start(void)
{
    ret_code_t err_code;

//    sd_ble_gap_adv_stop();
 //   sd_ble_gap_scan_stop();
    advertising_init();
    err_code = sd_ble_gap_adv_start(&m_adv_params, APP_BLE_CONN_CFG_TAG);
    APP_ERROR_CHECK(err_code);


    //err_code = bsp_indication_set(BSP_INDICATE_ADVERTISING);
    //APP_ERROR_CHECK(err_code);

}

/**@brief Function to start scanning.
 */
void scan_start(void)
{
    ret_code_t err_code;

//    sd_ble_gap_scan_stop();
//    sd_ble_gap_adv_stop();
    err_code = sd_ble_gap_scan_start(&m_scan_params);
    APP_ERROR_CHECK(err_code);
}

/**
 * @brief Parses advertisement data, providing length and location of the field in case
 *        matching data is found.
 *
 * @param[in]  type       Type of data to be looked for in advertisement data.
 * @param[in]  p_advdata  Advertisement report length and pointer to report.
 * @param[out] p_typedata If data type requested is found in the data report, type data length and
 *                        pointer to data will be populated here.
 *
 * @retval NRF_SUCCESS if the data type is found in the report.
 * @retval NRF_ERROR_NOT_FOUND if the data type could not be found.
 */
static uint32_t adv_report_parse(uint8_t type, data_t * p_advdata, data_t * p_typedata)
{
    uint32_t  index = 0;
    uint8_t * p_data;

    p_data = p_advdata->p_data;

    while (index < p_advdata->data_len)
    {
        uint8_t field_length = p_data[index];
        uint8_t field_type   = p_data[index + 1];

        if (field_type == type)
        {
            p_typedata->p_data   = &p_data[index + 2];
            p_typedata->data_len = field_length - 1;
            return NRF_SUCCESS;
        }
        index += field_length + 1;
    }
    return NRF_ERROR_NOT_FOUND;
}

static void handle_new_state(uint8_t old_relay_state, uint8_t new_relay_state)
{
    uint8_t old_state;
    uint8_t new_state;

    for(uint8_t index=0; index<6; index++)
    {
       old_state = old_relay_state & (1<<index);
       new_state = new_relay_state & (1<<index);
        if(old_state  != new_state)
        {
            if(new_state != 0)
            {
                if(index == 0)
                {
                    nrf_gpio_pin_set(RELAY_MAIN_1_PIO);
                }
                else if(index == 1)
                {
                    nrf_gpio_pin_set(RELAY_MAIN_2_PIO);
                }
                else if(index == 2)
                {
                    nrf_gpio_pin_set(RELAY_RIGHT_3_PIO);
                }
                else if(index == 3)
                {
                    nrf_gpio_pin_set(RELAY_RIGHT_4_PIO);
                }
                else if(index == 4)
                {
                    nrf_gpio_pin_set(RELAY_LEFT_5_PIO);
                }
                else if(index == 5)
                {
                    nrf_gpio_pin_set(RELAY_LEFT_6_PIO);
                }
            }
            else
            {
                if(index == 0)
                {
                    nrf_gpio_pin_clear(RELAY_MAIN_1_PIO);
                }
                else if(index == 1)
                {
                    nrf_gpio_pin_clear(RELAY_MAIN_2_PIO);
                }
                else if(index == 2)
                {
                    nrf_gpio_pin_clear(RELAY_RIGHT_3_PIO);
                }
                else if(index == 3)
                {
                    nrf_gpio_pin_clear(RELAY_RIGHT_4_PIO);
                }
                else if(index == 4)
                {
                    nrf_gpio_pin_clear(RELAY_LEFT_5_PIO);
                }
                else if(index == 5)
                {
                    nrf_gpio_pin_clear(RELAY_LEFT_6_PIO);
                }
            }
        }
    }
}

data_t     manu_data;
/**@brief Function for handling the advertising report BLE event.
 *
 * @param[in] p_ble_evt  Bluetooth stack event.
 */
static void on_adv_report(const ble_evt_t * const p_ble_evt)
{
    data_t     adv_data;
    ret_code_t err_code;

    // For readibility.
    ble_gap_evt_t  const * p_gap_evt  = &p_ble_evt->evt.gap_evt;
   // ble_gap_addr_t const * peer_addr  = &p_gap_evt->params.adv_report.peer_addr;

    // Initialize advertisement report for parsing
    adv_data.p_data   = (uint8_t *)p_gap_evt->params.adv_report.data;
    adv_data.data_len = p_gap_evt->params.adv_report.dlen;

    // Search for advertising names.
    err_code = adv_report_parse(BLE_GAP_AD_TYPE_MANUFACTURER_SPECIFIC_DATA, &adv_data, &manu_data);
            m_temp_scan_count++;

    if(err_code == NRF_SUCCESS && manu_data.data_len > 6)
    {
        /* Parse manufacure data to check the first 6 bytes including nordic as company identifier */
        if(manu_data.p_data[0] == 0x59 && manu_data.p_data[1] == 0x00 && manu_data.p_data[2] == 0x11 && manu_data.p_data[3] == 0x22 && 
           manu_data.p_data[4] == 0x33 && manu_data.p_data[5] == 0x44)
        {
            uint8_t old_relay_state = m_rbox_state.relay_state;

            /* This is Gear Knob beacon message */
            if(manu_data.p_data[6] == 0x01)
            {
                /* <sensor bitmask 1 byte>,<trigger switch 1 byte>,<Battery % 1 byte> */
                m_rbox_state.relay_state |= manu_data.p_data[7];
                m_rbox_state.trigger_sw_state = manu_data.p_data[8];
            }
            else if(manu_data.p_data[6] == 0x02)
            {
                /* sensor bitmask 1 byte */
                m_rbox_state.relay_state = manu_data.p_data[7];
            }
            else if(manu_data.p_data[6] == 0x03)
            {
                /* sensor value, Ignition off time */
                if(manu_data.p_data[7] <= 6)
                {
                    m_rbox_state.ig_off_time_relay[manu_data.p_data[7]] = uint16_decode(&manu_data.p_data[8]);
                }
            }
            handle_new_state(old_relay_state, m_rbox_state.relay_state);
        }
    }
}


/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
    // For readability.
   // ble_gap_evt_t const * p_gap_evt = &p_ble_evt->evt.gap_evt;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_ADV_REPORT:
        {
            on_adv_report(p_ble_evt);
        } break;

        default:
            // No implementation needed.
            break;
    }
}



/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}


/**@brief Function for initializing logging. */
static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/**@brief Function for initializing LEDs. */

static void leds_init(void)
{
    ret_code_t err_code = bsp_init(BSP_INIT_LED, NULL);
    APP_ERROR_CHECK(err_code);
}

static void scan_adv_timer_handler(void * p_context)
{
    UNUSED_PARAMETER(p_context);
    uint32_t timeout = SCAN_TIME;
    ret_code_t err_code;
    
    if(m_dev_state == dev_state_advert)
    {
        err_code = sd_ble_gap_adv_stop();
        APP_ERROR_CHECK(err_code);
        timeout = ADVERT_TIME_DELAY;
        m_dev_state = dev_state_advert_delay;
    }
    else if(m_dev_state == dev_state_advert_delay)
    {
        m_dev_state = dev_state_scan;
        scan_start();
    }
    else if(m_dev_state == dev_state_scan)
    {
        err_code = sd_ble_gap_scan_stop();
        APP_ERROR_CHECK(err_code);
        timeout = SCAN_TIME_DELAY;
        m_dev_state =dev_state_scan_delay;
    }
    else
    {
        timeout = ADVERT_TIME_DELAY;
        advertising_start();
        m_dev_state = dev_state_advert;
    }

    app_timer_stop(m_adv_timer_id);
    err_code = app_timer_start(m_adv_timer_id, timeout, NULL);
    APP_ERROR_CHECK(err_code);
}


/**@brief Function for initializing timers. */
static void timers_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);

    // Create timers.
    err_code = app_timer_create(&m_adv_timer_id,
                                APP_TIMER_MODE_SINGLE_SHOT,
                                scan_adv_timer_handler);
    APP_ERROR_CHECK(err_code);

}


/**@brief Function for doing power management.*/
static void power_manage(void)
{
    ret_code_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}


static void hardware_init(void)
{
    // Initialize led and turn off
    nrf_gpio_cfg_output(TEST_LED);
    nrf_gpio_pin_clear(TEST_LED);

    nrf_gpio_cfg(RELAY_MAIN_1_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);

    nrf_gpio_cfg(RELAY_MAIN_2_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);

    nrf_gpio_cfg(RELAY_RIGHT_3_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);

    nrf_gpio_cfg(RELAY_RIGHT_4_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);

    nrf_gpio_cfg(RELAY_LEFT_5_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);

    nrf_gpio_cfg(RELAY_LEFT_6_PIO, NRF_GPIO_PIN_DIR_OUTPUT, NRF_GPIO_PIN_INPUT_DISCONNECT,
                 NRF_GPIO_PIN_NOPULL, NRF_GPIO_PIN_S0H1 , NRF_GPIO_PIN_NOSENSE);
}

/**
 * @brief Function for application main entry.
 */
int main(void)
{
    //ret_code_t err_code;
    // Initialize.
    log_init();
    timers_init();
    leds_init();
    handle_new_state(1,0);

    ble_stack_init();
     // Hardware Initilize
    hardware_init();

    m_dev_state = dev_state_advert;
    advertising_start();
    scan_start();

/*
    app_timer_stop(m_adv_timer_id);
    err_code = app_timer_start(m_adv_timer_id, ADVERT_TIME, NULL);
    APP_ERROR_CHECK(err_code);
*/

    // Start execution.
    NRF_LOG_INFO("App Relay Control example started.");

    // Enter main loop.
    for (;; )
    {
        if (NRF_LOG_PROCESS() == false)
        {
            power_manage();
        }
    }
}


/**
 * @}
 */
